use crate::commands::backup::split_pointer_name;
use crate::descriptor::RetentionPolicyConfig;
use anyhow::{bail, Context};
use log::info;
use std::collections::{BTreeMap, BTreeSet};
use std::io;
use std::path::Path;
use yama::commands::open_pile;
use yama::operations::remove_pointer_safely;
use yama::pile::PileDescriptor;

pub struct RetentionBand {
    pub interval_s: u64,
    pub number_to_retain: u32,
}

pub struct RetentionPolicy {
    pub retention_bands: Vec<RetentionBand>,
}

const DAY: u64 = 86400;
const WEEK: u64 = 7 * DAY;
const MONTH: u64 = 31 * DAY;
const YEAR: u64 = 365 * DAY;

impl RetentionPolicy {
    pub fn from_config(descriptor: RetentionPolicyConfig) -> RetentionPolicy {
        let mut policy = RetentionPolicy {
            retention_bands: vec![],
        };

        if descriptor.daily != 0 {
            policy.retention_bands.push(RetentionBand {
                interval_s: DAY,
                number_to_retain: descriptor.daily,
            });
        }

        if descriptor.weekly != 0 {
            policy.retention_bands.push(RetentionBand {
                interval_s: WEEK,
                number_to_retain: descriptor.weekly,
            });
        }

        if descriptor.monthly != 0 {
            policy.retention_bands.push(RetentionBand {
                interval_s: MONTH,
                number_to_retain: descriptor.monthly,
            });
        }

        if descriptor.yearly != 0 {
            policy.retention_bands.push(RetentionBand {
                interval_s: YEAR,
                number_to_retain: descriptor.yearly,
            });
        }

        policy
    }

    /// Returns the set of snapshots to remove.
    pub fn apply_returning_prunable(
        &self,
        snapshots_by_unix_time: BTreeMap<u64, String>,
    ) -> BTreeSet<String> {
        if snapshots_by_unix_time.is_empty() {
            return BTreeSet::new();
        }
        let mut snapshots_included: BTreeSet<u64> = BTreeSet::new();

        // Always mark the most recent snapshot as retained!
        let last_snapshot = snapshots_by_unix_time.keys().rev().next().unwrap();
        snapshots_included.insert(*last_snapshot);

        let now_time = *last_snapshot;

        for band in &self.retention_bands {
            for multiple in 1..=band.number_to_retain {
                let target_time = now_time - (multiple as u64) * band.interval_s;
                if let Some((k, _)) = snapshots_by_unix_time.range(0..=target_time).rev().next() {
                    snapshots_included.insert(*k);
                }
            }
        }

        // Find all prunable (unincluded) snapshots.
        snapshots_by_unix_time
            .into_iter()
            .filter(|(k, _v)| !snapshots_included.contains(k))
            .map(|(_k, v)| v)
            .collect()
    }
}

pub fn prune_with_retention_policy(
    pile_path: &Path,
    pile_desc: &PileDescriptor,
    policy: &RetentionPolicy,
    prompt_first: bool,
) -> anyhow::Result<()> {
    let pile = open_pile(&pile_path, &pile_desc).context("Failed to open pile")?;

    let pointers = pile
        .list_pointers()
        .context("Failed to list pointers in pile")?;

    let mut pointers_to_keep: BTreeSet<String> = pointers.iter().cloned().collect();

    let pointers_to_remove = get_prunable_pointers(&policy, pointers);

    for remove in &pointers_to_remove {
        pointers_to_keep.remove(remove);
    }

    info!("Gory details:\n---\nKeep: {pointers_to_keep:?}\n---\nRemove: {pointers_to_remove:?}");
    info!(
        "{} pointers to remove ({} to keep) based on retention policy.",
        pointers_to_remove.len(),
        pointers_to_keep.len()
    );

    if prompt_first {
        println!("Would you like to proceed? [y/N]: ");
        let mut buffer = String::new();
        let stdin = io::stdin(); // We get `Stdin` here.
        stdin.read_line(&mut buffer)?;
        if buffer.trim().to_ascii_lowercase() != "y" {
            bail!("Aborted by user.");
        }
    }

    for to_remove in pointers_to_remove {
        let res = remove_pointer_safely(&pile, &to_remove).context("removing prunable pointers");

        pile.flush()
            .context("flushing pile after removing pointers")?;

        res?;
    }

    Ok(())
}

fn get_prunable_pointers(policy: &RetentionPolicy, pointers: Vec<String>) -> BTreeSet<String> {
    let mut split_pointers_by_name: BTreeMap<String, BTreeMap<u64, String>> = BTreeMap::new();

    for pointer in pointers {
        let (name, datetime) = if let Some(x) = split_pointer_name(&pointer) {
            x
        } else {
            continue;
        };

        split_pointers_by_name
            .entry(name)
            .or_default()
            .insert(datetime.timestamp().try_into().unwrap(), pointer);
    }

    let mut pointers_to_remove = BTreeSet::new();

    for (_pointer_base_name, ts_to_pointer) in split_pointers_by_name {
        let to_remove = policy.apply_returning_prunable(ts_to_pointer);

        pointers_to_remove.extend(to_remove);
    }

    pointers_to_remove
}

#[cfg(test)]
mod test {
    use crate::commands::prune::{get_prunable_pointers, RetentionPolicy};
    use crate::descriptor::RetentionPolicyConfig;

    #[test]
    fn test_prunable_pointers() {
        let pointers = vec![
            "alice+2022-09-28_05:00:00",
            "alice+2022-09-28_02:00:00",
            "alice+2022-09-21_05:00:00",
            "alice+2022-09-14_05:00:00",
            "alice+2022-09-08_05:00:00",
            "alice+2022-09-07_05:00:00",
            "alice+2022-09-01_05:00:00",
            "bob+2022-09-28_06:00:00",
            "bob+2022-09-28_03:00:00",
            "bob+2022-09-21_06:00:00",
            "bob+2022-09-14_06:00:00",
            "bob+2022-09-08_06:00:00",
            "bob+2022-09-07_06:00:00",
            "bob+2022-09-01_06:00:00",
        ]
        .into_iter()
        .map(|s| s.to_owned())
        .collect();
        let policy = RetentionPolicy::from_config(RetentionPolicyConfig {
            daily: 0,
            weekly: 3,
            monthly: 0,
            yearly: 0,
        });

        assert_eq!(
            get_prunable_pointers(&policy, pointers)
                .into_iter()
                .collect::<Vec<_>>(),
            vec![
                "alice+2022-09-01_05:00:00",
                "alice+2022-09-08_05:00:00",
                "alice+2022-09-28_02:00:00",
                "bob+2022-09-01_06:00:00",
                "bob+2022-09-08_06:00:00",
                "bob+2022-09-28_03:00:00",
            ]
        );
    }
}
