use crate::commands::{fully_integrate_pointer_node, retrieve_tree_node, store_tree_node};
use crate::pile::{Pile, RawPile};
use crate::tree::{differentiate_node_in_place, integrate_node_in_place};
use anyhow::{anyhow, Context};
use log::info;

pub mod checking;
pub mod cleanup;
pub mod extracting;
pub mod legacy_pushpull;
pub mod storing;

pub fn remove_pointer_safely<P: RawPile>(pile: &Pile<P>, name: &str) -> anyhow::Result<()> {
    // retrieve this pointer
    let mut this_pointer = pile
        .read_pointer(name)?
        .ok_or_else(|| anyhow!("Pointer {:?} does not exist so can not be deleted.", name))?;
    let mut this_node = retrieve_tree_node(&pile, this_pointer.chunk_ref.clone())
        .context("retrieving 'this' node")?;

    let new_parent_name = this_pointer.parent_pointer.clone();
    fully_integrate_pointer_node(pile, &mut this_node.node, &mut this_pointer)
        .context("integrating new parent")?;

    let new_parent = if let Some(ref new_parent_name) = new_parent_name {
        let mut new_parent_pointer = pile
            .read_pointer(new_parent_name.as_str())?
            .ok_or_else(|| anyhow!("Parent pointer {:?} does not exist.", name))?;
        let mut new_parent_node = retrieve_tree_node(&pile, new_parent_pointer.chunk_ref.clone())?;
        fully_integrate_pointer_node(pile, &mut new_parent_node.node, &mut new_parent_pointer)?;
        Some((new_parent_pointer, new_parent_node))
    } else {
        None
    };

    // now integrate any pointers that rely on this one
    // so that they no longer rely on this one.
    for pointer in pile.list_pointers()?.iter() {
        if pointer == name {
            continue;
        }
        if let Some(mut pointer_data) = pile.read_pointer(pointer.as_str())? {
            if let Some(parent_pointer) = pointer_data.parent_pointer.as_ref() {
                if parent_pointer == name {
                    info!("Pointer would be orphaned: {:?}; integrating", pointer);

                    // need to integrate this node, so retrieve it
                    let mut node = retrieve_tree_node(&pile, pointer_data.chunk_ref)?;

                    // integrate it in-place
                    integrate_node_in_place(&mut node.node, &this_node.node)?;

                    if let Some((_, ref new_parent_node)) = new_parent {
                        // then differentiate with respect to the NEW parent
                        differentiate_node_in_place(&mut node.node, &new_parent_node.node)?;
                    }

                    // pass through the parent
                    pointer_data.parent_pointer = new_parent_name.clone();

                    // store the updated version of the pointer
                    let new_chunk_ref = store_tree_node(&pile, &node)?;
                    // associate the new node with the new version of the pointer
                    pointer_data.chunk_ref = new_chunk_ref;
                    // write the pointer back.
                    pile.write_pointer(pointer.as_str(), &pointer_data)?;

                    // we must flush chunks before deleting the pointer
                    pile.flush()
                        .context("flushing after writing pointer back")?;
                }
            }
        }
    }

    // then delete the pointer
    pile.delete_pointer(name)?;
    info!("Deleted pointer: {:?}", name);
    Ok(())
}
