use crate::pile::local_sqlitebloblogs::{CompactionThresholds, SqliteBloblogPile};
use crate::pile::{PileDescriptor, PileStorage};
use anyhow::{bail, Context};
use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
use log::info;
use std::path::Path;

pub fn compact(
    pile_path: &Path,
    pile_desc: &PileDescriptor,
    actually_run: bool,
    make_progress_bar: bool,
    thresholds: CompactionThresholds,
) -> anyhow::Result<()> {
    let pbar = if make_progress_bar {
        ProgressBar::with_draw_target(1000 as u64, ProgressDrawTarget::stdout_with_hz(10))
    } else {
        ProgressBar::hidden()
    };
    pbar.set_style(
        ProgressStyle::default_bar()
            .template("[{elapsed_precise}]/[{eta}] {bar:40.cyan/blue} {pos:>7}/{len:7} {msg}"),
    );
    pbar.set_message("compacting");

    match pile_desc.storage {
        PileStorage::SqliteIndexedBloblog => {
            let bloblog_pile = SqliteBloblogPile::open(&pile_path)
                .context("Failed to open SQLite-indexed Bloblog Pile")?;
            compact_bloblogs(bloblog_pile, pbar, actually_run, thresholds)?;
            Ok(())
        }
        other @ PileStorage::RemoteOnly => {
            bail!("Cannot use compaction on this kind of pile: {other:?}!");
        }
    }
}

fn compact_bloblogs(
    bloblog_pile: SqliteBloblogPile,
    pbar: ProgressBar,
    actually_run: bool,
    thresholds: CompactionThresholds,
) -> anyhow::Result<()> {
    info!("=== Analysing for compaction ===");
    let analysis = bloblog_pile.analyse_for_compaction()?;
    let chunks_total: u64 = analysis.values().map(|bs| bs.chunks_total).sum();
    let chunks_deleted: u64 = analysis.values().map(|bs| bs.chunks_deleted).sum();
    let bytes_total: u64 = analysis.values().map(|bs| bs.bytes_total).sum();
    let bytes_deleted: u64 = analysis.values().map(|bs| bs.bytes_deleted).sum();

    info!("{} bloblogs in this pile, with {chunks_total} chunks ({bytes_total} B) of which {chunks_deleted} ({bytes_deleted} B) are deleted.", analysis.len());

    info!("=== Planning compaction ===");
    let plan = bloblog_pile.plan_compaction(&thresholds, analysis)?;
    info!("Planned compaction: replace {} bloblogs (of which {} are small), freeing up {} B and rewriting {} B", plan.bloblogs_to_replace.len(), plan.small_bloblogs, plan.reclaimable_space, plan.bytes_to_write);

    if actually_run {
        info!("=== Compacting ===");
        bloblog_pile.perform_compaction(Box::new(pbar), plan)?;
    }

    Ok(())
}
